/*
 * =====================================================================================
 *
 *       Filename:  bch.c
 *
 *    Description:  systematic BCH encoder and decoder implementation
 *
 *        Version:  1.0
 *        Created:  2020-06-02 03:40:39 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Alvin Sukmadji 
 *   Organization:  University of Toronto
 *   
 *   Note: This license is a modified BSD License, which contains the terms of noncommercial
 *   use.
 *   
 *   Copyright <2020> <Alvin Yonathan Sukmadji and Frank Robert Kschischang>
 *   
 *   Redistribution and use in source and binary forms, with or without modification, are
 *   permitted provided that the following conditions are met:
 *   
 *   1. Redistributions of source code must retain the above copyright notice, this list of
 *   conditions and the following disclaimer.
 *   
 *   2. Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *   
 *   3. Neither the name of the copyright holder nor the names of its contributors may be used
 *   to endorse or promote products derived from this software without specific prior written
 *   permission.
 *   
 *   4. This software may only be used for noncommercial purposes, including teaching and
 *   academic research at universities, colleges and other educational institutions, and
 *   personal non-profit purposes. Neither this software, whether in Source or Object form,
 *   nor any derivative work that is based on (or derived from) this software may be used for
 *   commercial purposes, for example, licensing or selling the software, or using the
 *   software with a purpose to procure a commercial gain.
 *   
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <time.h>
#include "pcg_variants.h"
#include "tables128.h" // UNCOMMENT THIS LINE TO USE GF(128) FOR THE ERROR-LOCATOR FIELD
// #include "tables256.h" // UNCOMMENT THIS LINE TO USE GF(256) FOR THE ERROR-LOCATOR FIELD
// #include "tables512.h" // UNCOMMENT THIS LINE TO USE GF(512) FOR THE ERROR-LOCATOR FIELD
// #include "tables1024.h" // UNCOMMENT THIS LINE TO USE GF(1024) FOR THE ERROR-LOCATOR FIELD
#include "bch.h"

#define N GFq
#define K (N - 3*GFm)

/**************************** GENERATOR POLYNOMIALS - UNCOMMENT THE APPROPRIATE ONE BASED ON THE FIELD SIZE ****************************/
/* for example, if using GF(128), uncomment the generator polynomial for (127,106) BCH */

// generator polynomial coef for (127,106) BCH with d=7 - for use with tables128.h
// x^21 + x^19 + x^16 + x^13 + x^12 + x^4 + x^3 + x^1 + x^0
static const unsigned char genpoly[N-K+1] = {1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1};

// generator polynomial coef for (255,231) BCH with d=7 - for use with tables256.h
// x^24 + x^23 + x^21 + x^20 + x^19 + x^17 + x^16 + x^15 + x^13 + x^8 + x^7 + x^5 + x^4 + x^2 + x^0
// static const unsigned char genpoly[N-K+1] = {1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1};

// generator polynomial coef for (511,484) BCH with d=7 - for use with tables512.h
// x^27 + x^26 + x^24 + x^22 + x^21 + x^16 + x^13 + x^11 + x^9 + x^8 + x^6 + x^5 + x^4 + x^3 + x^0
// static const unsigned char genpoly[N-K+1] = {1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1};

// generator polynomial coef for (1023,993) BCH with d=7 - for use with tables1024.h
// x^30 + x^28 + x^23 + x^21 + x^19 + x^16 + x^12 + x^8 + x^4 + x^1 + x^0
// static const unsigned char genpoly[N-K+1] = {1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1};

/***************************************************************************************************************************************/

/* fill a buffer of unsigned chars with random {0,1} elements  */
void fillrand(unsigned char *buffer, int len, pcg32f_random_t *rng)
{
   int i,j;
   uint32_t randword;

   j = 0;
   for (i=0; i<len; ++i)
   {
      if (j < 1)
      {
         randword = pcg32f_random_r(rng);
         j = 32;
      }
      buffer[i] = randword & 0x01;
      randword >>= 1;
      j--;
   }
}

// prints BCH codeword, for debugging
void BCH_print(unsigned char *x, int n)
{
	for (int i = 0; i < n; i++)
	{
		printf("%u", x[i]);
	}
	printf("\n");
}

// rem size n-k
// computing parity bits of x array, uses long division
// writes the remainder in rem
void compute_parity(unsigned char *x, unsigned char *rem, int n, int k)
{
	int buflen = n-k+1;
	// set the last n-k entries of x to zero
	for(int i = k; i < n; i++)
	{
		x[i] &= 0x00;
	}
	// BCH_print(x,n);

	// long division
	for(int i = 0; i < k; i++)
	{
		if(x[i] != 0x00)
		{
			for(int j = 0; j < buflen; j++)
			{
				x[i+j] ^= genpoly[j];
			}
		}
	}

	// copy remainder to rem
	for(int i = 0; i < n-k; i++)
	{
		rem[i] = x[k+i];
	}
}

// long division of x with genpoly over GF(2)
// writes the remainder in rem
void long_division(unsigned char *x, unsigned char *rem, int n, int k)
{
   int buflen = n-k+1;
   for(int i = 0; i < k; i++)
   {
      if(x[i] != 0x00)
      {
         for(int j = 0; j < buflen; j++)
         {
            x[i+j] ^= genpoly[j];
         }
      }
   }

   // copy remainder to rem
   for(int i = 0; i < n-k; i++)
   {
      rem[i] = x[k+i];
   }
}

// systematic BCH encoder implementation
void BCH_encode(unsigned char *x, int n, int k)
{
	// set the last n-k entries of x to zero
	for(int i = k; i < n; i++)
	{
		x[i] &= 0x0;
	}

	// set copy of x for long division
	int buflen = n-k+1;
	unsigned char buf[n];

	for(int i = 0; i < n; i++)
	{
		buf[i] = x[i];
	}

	// long division
	for(int i = 0; i < k; i++)
	{
		if(buf[i] != 0)
		{
			for(int j = 0; j < buflen; j++)
			{
				buf[i+j] ^= genpoly[j];
			}
		}
	}

	// copy remainder to x
	for(int i = k; i < n; i++)
	{
		x[i] = buf[i];
	}
}

// BCH decoder implementation
int BCH_decode(unsigned long long int syndrome,
              unsigned int *p1, unsigned int *p2, unsigned int *p3)
{
   GFel S1, S3, S5, D3, D5, r1, r2, b, c;



   if (!(syndrome & 0x3FFFFFFFFFFFFFFF)) 
      return 0;                   /* t = 0 */

   /* Unpack the syndrome */
   S5 = (GFel) (syndrome & BITMASK);
   syndrome >>= GFm; S3 = (GFel) (syndrome & BITMASK);
   syndrome >>= GFm; S1 = (GFel) (syndrome & BITMASK);

   /* Compute discrepancies */
   D3 = PLUS(Pow3[S1],S3);
   D5 = PLUS(Pow5[S1],S5);

   if (S1 && !(D3) && !(D5)) /*  t = 1 */
   {
      *p1 = LOG[S1]; return 1;
   }

   if (S1 && D3 && !(PLUS(TIMES(S1,D5),TIMES(S3,D3)))) /* t=2 */
   {
      c = TIMES(D3,Reciprocal[Pow3[S1]]) << 1;
      if ((*p1 = TIMES(S1, DepQuadRoot[c]))) /* if we have a valid error */
      {
      	 *p1 = LOG[*p1];
         *p2 = LOG[TIMES(S1, DepQuadRoot[c+1])];
         return 2;
      }
      else
         return -1;
   }

   if (D3) /* t = 3 */
   {
      if (D5)  /* if D5 is nonzero, we have a depressed cubic */
      {
         c = SquareRoot[TIMES(Pow5[D3],Reciprocal[Pow3[D5]])] << 1;
         b = SquareRoot[TIMES(D5,Reciprocal[D3])];
         r1 = DepCubicRoot[c];
         if (r1)
         {
            r2 = DepCubicRoot[c+1];
            *p1 = PLUS(S1,TIMES(b,r1));
            *p2 = PLUS(S1,TIMES(b,r2));
            *p3 = LOG[PLUS(S1,PLUS(*p1,*p2))];
            *p1 = LOG[*p1];
            *p2 = LOG[*p2];
            return 3;
         }
         else
            return -1;
      }
      else /* if D5 is zero, we have a pure cubic */
      {
         c = D3 << 1;
         r1 = CubeRoot[c];
         if (r1)
         {
            r2 = CubeRoot[c+1];
            *p1 = LOG[PLUS(S1,r1)];
            *p2 = LOG[PLUS(S1,r2)];
            *p3 = LOG[PLUS(S1,PLUS(r1,r2))];
            return 3;
         }
         else
            return -1;
      }
   }
   /* default --- uncorrectable */
   return -1;
}

// computing the syndrome of a BCH codeword x of length n<=4095
unsigned long long int BCH_compute_syndrome(unsigned char *x, int n)
{
	unsigned long long int s;
	unsigned long long int syndrome = 0x0;
	for(int i = 0; i < n; i++)
	{
		if(x[i])
		{
			s = EXP[(n-i-1) % GFq];
			// s = (n-i-1) % GFq;
			syndrome ^= ((s)<<2*GFm)^(Pow3[s]<<GFm)^(Pow5[s]);
		}
	}
	return syndrome;
}

// compute the syndrome of a BCH codeword with 1 at position pos and 0 everywhere else, n<=4095
unsigned long long int BCH_compute_syndrome_pos(unsigned int pos, int n)
{
	unsigned long long int s = EXP[(n-pos-1) % GFq];
	unsigned long long int syndrome = 0x0;
	syndrome ^= ((s)<<2*GFm)^(Pow3[s]<<GFm)^(Pow5[s]);
	return syndrome;
}

// get BCH code length
int BCH_get_n()
{
   return N;
}

// get BCH code dimension
int BCH_get_k()
{
   return K;
}


























