/*
 * =====================================================================================
 *
 *       Filename:  common.h
 *
 *    Description:  implementations for encoding and decoding of zipper codes
 *
 *        Version:  1.0
 *        Created:  2020-11-29 08:23:44 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Alvin Sukmadji 
 *   Organization:  University of Toronto
 *   
 *   Note: This license is a modified BSD License, which contains the terms of noncommercial
 *   use.
 *   
 *   Copyright <2020> <Alvin Yonathan Sukmadji>
 *   
 *   Redistribution and use in source and binary forms, with or without modification, are
 *   permitted provided that the following conditions are met:
 *   
 *   1. Redistributions of source code must retain the above copyright notice, this list of
 *   conditions and the following disclaimer.
 *   
 *   2. Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *   
 *   3. Neither the name of the copyright holder nor the names of its contributors may be used
 *   to endorse or promote products derived from this software without specific prior written
 *   permission.
 *   
 *   4. This software may only be used for noncommercial purposes, including teaching and
 *   academic research at universities, colleges and other educational institutions, and
 *   personal non-profit purposes. Neither this software, whether in Source or Object form,
 *   nor any derivative work that is based on (or derived from) this software may be used for
 *   commercial purposes, for example, licensing or selling the software, or using the
 *   software with a purpose to procure a commercial gain.
 *   
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * =====================================================================================
 */

#include "zipper.h"
#include "pcg_variants.h"
#include "bch.h"

#define MAX_ITER 10 // maximum allowed number of iteration for iterative decoding

// custom parser for zipper code parameter
static void parse_file(ZIPPER *zipper, char *filename);

// setting up parameters of zipper code by parsing the configuration file
// and allocating memory for the zipper struct
// inputs:
//     zipper    : pointer to the zipper code struct
//     filename  : filename (string) of the configuration file
//     chunk_size: the number of rows in a chunk, i.e., 'packet'
void zipper_setup(ZIPPER *zipper, char *filename, int chunk_size)
{
	if(chunk_size <= 0)
	{
		printf("ERROR: chunk_size must be a positive integer\n");
		exit(1);
	}
	zipper->chunk_size = chunk_size;
	zipper->max_lookback = -1;
	zipper->n_max = -1;
	parse_file(zipper, filename);
	zipper->extended_buffer_len = zipper->buffer_len + 2*(zipper->chunk_size + zipper->max_lookback);

	// allocate memory for buffer
	zipper->buffer_sent = malloc(zipper->extended_buffer_len * sizeof(char*));
	zipper->buffer_recv = malloc(zipper->extended_buffer_len * sizeof(char*));
	zipper->syndromes = malloc(zipper->extended_buffer_len * sizeof(unsigned long long int));
	zipper->is_fresh = malloc(zipper->extended_buffer_len * sizeof(bool));
	for(int i = 0; i < zipper->extended_buffer_len; i++)
	{
		zipper->buffer_sent[i] = malloc(zipper->n_max * sizeof(char));
		zipper->buffer_recv[i] = malloc(zipper->n_max * sizeof(char));
		for(int j = 0; j < zipper->n_max; j++)
		{
			zipper->buffer_sent[i][j] &= 0x00;
			zipper->buffer_recv[i][j] &= 0x00;
		}
		zipper->syndromes[i] &= 0x00;
	}
}

// encoding the latest chunk and transmitting it to the receiver
// assumes transmission over the binary symmetric channel with crossover probability error_prob
// inputs:
//     zipper          : pointer to the zipper code struct
//     newest_chunk_id : number of chunks transmitted so far
//     rng             : pointer to the ZMG (or PCG) random number generator
//     error_prob      : crossover probability over the BSC 
void zipper_encode(ZIPPER *zipper, long long int newest_chunk_id, pcg32f_random_t *rng, double error_prob)
{
	long long int chunk_row_start = newest_chunk_id * zipper->chunk_size;
	int row_start = (int) chunk_row_start % zipper->extended_buffer_len;
	int period_start = (int) chunk_row_start % zipper->period;

	// copy old symbols to virtual buffer and add new information symbols to real buffer
	for(int i = 0; i < zipper->chunk_size; i++)
	{
		int row_id = (row_start+i) % zipper->extended_buffer_len;
		int period_id = (period_start+i) % zipper->period;
		for(int j = 0; j < zipper->m[period_id]; j++)
		{
			int old_row = (row_id + zipper->i_map[period_id][j]->row_diff + zipper->extended_buffer_len) % zipper->extended_buffer_len;
			zipper->buffer_sent[row_id][j] = zipper->buffer_sent[old_row][zipper->i_map[period_id][j]->col];
		}
		fillrand(&zipper->buffer_sent[row_id][zipper->m[period_id]], zipper->k[period_id]-zipper->m[period_id], rng);
		BCH_encode(zipper->buffer_sent[row_id], zipper->n[period_id], zipper->k[period_id]);
	}

	// fill in receiver side buffer
	for(int i = 0; i < zipper->chunk_size; i++)
	{
		int row_id = (row_start+i) % zipper->extended_buffer_len;
		int period_id = (period_start+i) % zipper->period;

		// copy older symbol to virtual buffer
		for(int j = 0; j < zipper->m[period_id]; j++)
		{
			int old_row = (row_id + zipper->i_map[period_id][j]->row_diff + zipper->extended_buffer_len) % zipper->extended_buffer_len;
			zipper->buffer_recv[row_id][j] = zipper->buffer_recv[old_row][zipper->i_map[period_id][j]->col];
		}

		// copy symbol from transmitter and introduce errors
		for(int j = zipper->m[period_id]; j < zipper->n[period_id]; j++)
		{
			zipper->buffer_recv[row_id][j] = zipper->buffer_sent[row_id][j];
			double d = ldexp(pcg32f_random_r(rng), -32);
			if(d < error_prob)
			{
				zipper->buffer_recv[row_id][j] ^= 0x01;
			}
			
		}
		zipper->syndromes[row_id] = BCH_compute_syndrome(zipper->buffer_recv[row_id],zipper->n[period_id]);
		zipper->is_fresh[row_id] = (zipper->syndromes[row_id] != 0x00); // set fresh flag for newly received rows with nonzero syndrome
	}
}

static int get_error_locations(unsigned long long int syndrome, unsigned int *error_locations, int n)
{
	int num_error = BCH_decode(syndrome, error_locations, error_locations+1, error_locations+2);
	for(int i = 0; i < num_error; i++)
	{
		if(error_locations[i] >= n) // error located on bits that are not sent
		{
			return -1;
		}
	}
	return num_error;
}

// decoding codewords stored in the decoding window via iterative decoding
// returns the number of erroneous symbols in the oldest chunk of the window
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
//     num_info_symbols : pointer to the number of information symbols collected so far
unsigned int zipper_decode(ZIPPER *zipper, long long int newest_chunk_id, long long int *num_info_symbols)
{
	long long int last_row = newest_chunk_id * zipper->chunk_size + zipper->chunk_size - 1;
	int row_end = (int) last_row % zipper->extended_buffer_len;
	int period_end = (int) last_row % zipper->period;

	int iter = 0;
	bool modified = true;
	while(modified && iter < MAX_ITER)
	{
		modified = false;
		unsigned int p[3]; // error locations
		for(int i = 0; i < zipper->buffer_len; i++)
		{
			int row_id = (row_end - i + zipper->extended_buffer_len) % zipper->extended_buffer_len;
			int period_id = (period_end - i%zipper->period + zipper->period) % zipper->period;
			if(zipper->is_fresh[row_id])
			{
				int num_errors = get_error_locations(zipper->syndromes[row_id], p, zipper->n[period_id]);
				for(int j = 0; j < num_errors; j++)
				{
					unsigned int pos = zipper->n[period_id] - p[j] - 1;
					if(pos < zipper->m[period_id])
					{
						long long int real_row = last_row - i + zipper->i_map[period_id][pos]->row_diff;
						if(real_row < 0)
						{
							num_errors = -1; //miscorrection due to flipping bit on row index < 0
						}
					}
				}
				if(num_errors > 0)
				{
					for(int j = 0; j < num_errors; j++)
					{
						unsigned int pos = zipper->n[period_id] - p[j] - 1;
						if(pos < zipper->m[period_id]) // error in the virtual buffer
						{
							long long int real_row = last_row - i + zipper->i_map[period_id][pos]->row_diff;
							assert(real_row >= 0);
							int real_row_id = (int) real_row % zipper->extended_buffer_len;
							int real_period_id = (int) real_row % zipper->period;
							int real_pos = zipper->i_map[period_id][pos]->col;
							for(int k = 0; k < zipper->num_copy[real_period_id][real_pos]; k++)
							{
								long long int virt_row = real_row + zipper->i_map[real_period_id][real_pos][k].row_diff;
								int virt_row_id = (int) virt_row % zipper->extended_buffer_len;
								int virt_period_id = (int) virt_row % zipper->period;
								int virt_pos = zipper->i_map[real_period_id][real_pos][k].col;
								zipper->syndromes[virt_row_id] ^= BCH_compute_syndrome_pos(virt_pos, zipper->n[virt_period_id]);
								zipper->buffer_recv[virt_row_id][virt_pos] ^= 0x01;
								zipper->is_fresh[virt_row_id] = true;
							}
							zipper->buffer_recv[real_row_id][real_pos] ^= 0x01;
							zipper->syndromes[real_row_id] ^= BCH_compute_syndrome_pos(real_pos, zipper->n[real_period_id]);
							zipper->is_fresh[real_row_id] = true;
						}
						else // error in the real buffer
						{
							for(int k = 0; k < zipper->num_copy[period_id][pos]; k++)
							{
								long long int virt_row = last_row - i + zipper->i_map[period_id][pos][k].row_diff;
								if(virt_row <= last_row) // ignore if the virtual copy is out of the decoding window range
								{
									int virt_row_id = (int) virt_row % zipper->extended_buffer_len;
									int virt_period_id = (int) virt_row % zipper->period;
									int virt_pos = zipper->i_map[period_id][pos][k].col;
									zipper->syndromes[virt_row_id] ^= BCH_compute_syndrome_pos(virt_pos, zipper->n[virt_period_id]);
									zipper->buffer_recv[virt_row_id][virt_pos] ^= 0x01;
									zipper->is_fresh[virt_row_id] = true;
								}
							}
							zipper->buffer_recv[row_id][pos] ^= 0x01;
							zipper->syndromes[row_id] ^= BCH_compute_syndrome_pos(pos, zipper->n[period_id]);
						}
					}
					zipper->is_fresh[row_id] = false;
					modified = true;
				}
				else // either no error or decoding failure
				{
					zipper->is_fresh[row_id] = false;
				}
			}
		}
		iter++;
	}

	//count number of error in the oldest chunk
	long long int oldest_row = last_row - zipper->buffer_len + 1;
	if(oldest_row < 0)
	{
		return 0;
	}
	int errcount = 0;
	for(int i = 0; i < zipper->chunk_size; i++)
	{
		int row_id = (int) (oldest_row + i) % zipper->extended_buffer_len;
		int period_id = (int) (oldest_row + i) % zipper->period;
		for(int j = zipper->m[period_id]; j < zipper->k[period_id]; j++)
		{
			errcount += (zipper->buffer_sent[row_id][j] ^ zipper->buffer_recv[row_id][j]);
		}
		*num_info_symbols += (zipper->k[period_id] - zipper->m[period_id]);
	}

	return errcount;
}

// frees up memory allocated in zipper_setup
// inputs:
//     zipper : pointer to the zipper code struct
void zipper_cleanup(ZIPPER *zipper)
{
	// free interleaver map memory
	for(int i = 0; i < zipper->period; i++)
	{
		for(int j = 0; j < zipper->n[i]; j++)
		{
			if(zipper->i_map[i][j] != NULL)
			{
				free(zipper->i_map[i][j]);
			}
		}
		free(zipper->i_map[i]);
		free(zipper->num_copy[i]);
	}
	free(zipper->i_map);
	free(zipper->num_copy);
	free(zipper->m);
	free(zipper->n);
	free(zipper->k);

	//free buffer memory
	for(int i = 0; i < zipper->extended_buffer_len; i++)
	{
		free(zipper->buffer_sent[i]);
		free(zipper->buffer_recv[i]);
	}
	free(zipper->buffer_sent);
	free(zipper->buffer_recv);
	free(zipper->syndromes);
	free(zipper->is_fresh);
}

// prints the symbols in the transmitter-side buffer
// the number of bits printed out will be as many as the number of bits in the
// decoding window, with chunk no. newest_chunk_id being the newest (i.e., last) one
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
void zipper_print_buffer_sent(ZIPPER *zipper, long long int newest_chunk_id)
{
	long long int chunk_row_start = newest_chunk_id * zipper->chunk_size;
	int row_start = (int) (chunk_row_start + zipper->chunk_size - zipper->buffer_len + zipper->extended_buffer_len) % zipper->extended_buffer_len;
	int period_start = (int) ((chunk_row_start % zipper->period) - (zipper->buffer_len - zipper->chunk_size)%zipper->period + zipper->period) % zipper->period;
	for(int i = 0; i < zipper->buffer_len; i++)
	{
		int row_id = (row_start + i) % zipper->extended_buffer_len;
		int period_id = (period_start + i) % zipper->period;
		printf("row %5d: ", row_id);
		for(int j = 0; j < zipper->n[period_id]; j++)
		{
			printf("%u", zipper->buffer_sent[row_id][j]);
		}
		printf("\n");
	}
}

// prints the symbols in the receiver-side buffer
// the number of bits printed out will be as many as the number of bits in the
// decoding window, with chunk no. newest_chunk_id being the newest (i.e., last) one
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
void zipper_print_buffer_recv(ZIPPER *zipper, long long int newest_chunk_id)
{
	long long int chunk_row_start = newest_chunk_id * zipper->chunk_size;
	int row_start = (int) (chunk_row_start + zipper->chunk_size - zipper->buffer_len + zipper->extended_buffer_len) % zipper->extended_buffer_len;
	int period_start = (int) ((chunk_row_start % zipper->period) - (zipper->buffer_len-zipper->chunk_size)%zipper->period + zipper->period) % zipper->period;
	for(int i = 0; i < zipper->buffer_len; i++)
	{
		int row_id = (row_start + i) % zipper->extended_buffer_len;
		int period_id = (period_start + i) % zipper->period;
		printf("row %5d: ", row_id);
		for(int j = 0; j < zipper->n[period_id]; j++)
		{
			printf("%u", zipper->buffer_recv[row_id][j]);
		}
		printf("\n");
	}
}

static void parse_file(ZIPPER *zipper, char *filename)
{
	if(access(filename, F_OK) == -1)
	{
		printf("ERROR: filename %s does not exist\n", filename);
		exit(1);
	}
	FILE* file = fopen(filename, "r");
	char line[4000000];
	char *param;
	char delims[5] = " ,\n\0";
	bool declared[6] = {false, false, false, false, false, false}; // period, n, k, m, interleaver map
	while(fgets(line,sizeof(line),file))
	{
		if((line[0] == '/' && line[1] == '/') || line[0] == '\0' || line[0] == '\n') //comment or empty line
		{
			continue;
		}
		param = strtok(line, delims);
		if(strcmp(param, "period") == 0)
		{
			zipper->period = atoi(strtok(NULL, delims));
			if(zipper->period <= 0)
			{
				printf("ERROR: period has to be a positive integer\n");
				exit(1);
			}
			zipper->m = malloc(zipper->period * sizeof(int));
			zipper->n = malloc(zipper->period * sizeof(int));
			zipper->k = malloc(zipper->period * sizeof(int));
			declared[0] = true;
		}
		else if(strcmp(param, "n") == 0) //parse n
		{
			if(!declared[0]) // period needs to be declared first
			{
				printf("ERROR: the period of the zipper code needs to be declared first\n");
				exit(1);
			}
			for(int i = 0; i < zipper->period; i++)
			{
				char *val = strtok(NULL, delims);
				if(val == NULL)
				{
					printf("ERROR: the number of n's declared has to be the same as the period\n");
					exit(1);
				}
				zipper->n[i] = atoi(val);
				if(zipper->n[i] < 0 || zipper->n[i] > BCH_get_n())
				{
					printf("ERROR: each n[i] has to be a nonnegative integer less than or equal to the constituent code length %d\n", BCH_get_n());
					exit(1);
				}
				if(zipper->n[i] > zipper->n_max)
				{
					zipper->n_max = zipper->n[i];
				}
			}
			declared[1] = true;
		}
		else if(strcmp(param, "k") == 0) //parse k
		{
			if(!(declared[0] && declared[1]))
			{
				printf("ERROR: the period and n of the zipper code need to be declared first\n");
				exit(1);
			}
			for(int i = 0; i < zipper->period; i++)
			{
				char *val = strtok(NULL, delims);
				if(val == NULL)
				{
					printf("ERROR: the number of k's declared has to be the same as the period\n");
					exit(1);
				}
				zipper->k[i] = atoi(val);
				if(zipper->k[i] >= zipper->n[i] || zipper->k[i] < 0 || zipper->n[i]-zipper->k[i] < BCH_get_n()-BCH_get_k())
				{
					printf("ERROR: the value of each k[i] has to be a nonnegative integer lower than n[i], and n[i]-k[i] must be at least %d (the number of parity symbols of the constituent BCH code) \n", BCH_get_n()-BCH_get_k());
					exit(1);
				}
			}
			declared[2] = true;
		}
		else if(strcmp(param, "m") == 0) //parse m
		{
			if(!(declared[0] && declared[1] && declared[2]))
			{
				printf("ERROR: the period, n, and k of the zipper code need to be declared first\n");
				exit(1);
			}
			for(int i = 0; i < zipper->period; i++)
			{
				char *val = strtok(NULL, delims);
				if(val == NULL)
				{
					printf("ERROR: the number of k's declared has to be the same as the period\n");
					exit(1);
				}
				zipper->m[i] = atoi(val);
				if(zipper->k[i] < zipper->m[i] || zipper->m[i] < 0)
				{
					printf("ERROR: the value of each m[i] has to be a nonnegative integer of at most k[i]\n");
					exit(1);
				}
			}
			declared[3] = true;
		}
		else if(strcmp(param, "buffer_len") == 0) // parse maximum allowable iteration
		{
			zipper->buffer_len = atoi(strtok(NULL, delims));
			if(zipper->buffer_len <= 0)
			{
				printf("ERROR: buffer_len has to be a positive integer\n");
				exit(1);
			}
			declared[5] = true;
		}
		else if(strcmp(param, "interleaver_map") == 0) // parse interleaver map
		{
			if(!(declared[0] && declared[1] && declared[2] && declared[3]))
			{
				printf("ERROR: the period, n, k, and m of the zipper code need to be declared first\n");
				exit(1);
			}
			zipper->i_map = malloc(zipper->period * sizeof(I_MAP**));
			zipper->num_copy = malloc(zipper->period * sizeof(int*));
			for(int i = 0; i < zipper->period; i++)
			{
				zipper->i_map[i] = malloc(zipper->n[i] * sizeof(I_MAP*));
				zipper->num_copy[i] = malloc(zipper->n[i] * sizeof(int));
				for(int j = 0; j < zipper->m[i]; j++)
				{
					zipper->i_map[i][j] = malloc(sizeof(I_MAP));
					zipper->num_copy[i][j] = 1;
				}
				for(int j = zipper->m[i]; j < zipper->n[i]; j++)
				{
					zipper->i_map[i][j] = NULL;
					zipper->num_copy[i][j] = 0;
				}
			}

			for(int i = 0; i < zipper->period; i++)
			{
				char* curr_line = fgets(line,sizeof(line),file);
				int row_diff, col;
				if(curr_line == NULL && zipper->m[i] > 0)
				{
					printf("ERROR: the interleaver map for line %d is missing\n", i);
					exit(1);
				}
				row_diff = atoi(strtok(line, delims));
				for(int j = 0; j < zipper->m[i]; j++)
				{
					col = atoi(strtok(NULL, delims));
					if(row_diff <= 0)
					{
						printf("ERROR: row_diff[%d][%d] has to be a positive integer\n", i, j);
						exit(1);
					}
					int row_period = (i - row_diff%zipper->period + zipper->period) % zipper->period;
					if(col < zipper->m[row_period] || col >= zipper->n[row_period])
					{
						printf("ERROR: col[%d][%d]=%d has to be an integer greater than %d but lower than %d\n", i, j, col, zipper->m[row_period], zipper->n[row_period]);
						exit(1);
					}
					if(row_diff > zipper->max_lookback)
					{
						zipper->max_lookback = row_diff;
					}
					zipper->i_map[i][j][0].row_diff = -1 * row_diff;
					zipper->i_map[i][j][0].col = col;
					fflush(stdout);
					//need to resize interleaver map entry based on number of copies for real buffer
					int row_in_real = (i - (row_diff%zipper->period) + zipper->period) % zipper->period;
					I_MAP *new_map = malloc((zipper->num_copy[row_in_real][col]+1)*sizeof(I_MAP));
					for(int k = 0; k < zipper->num_copy[row_in_real][col]; k++)
					{
						new_map[k] = zipper->i_map[row_in_real][col][k];
					}
					new_map[zipper->num_copy[row_in_real][col]].row_diff = row_diff;
					new_map[zipper->num_copy[row_in_real][col]].col = j;
					if(zipper->i_map[row_in_real][col] != NULL)
					{
						free(zipper->i_map[row_in_real][col]);
					}
					fflush(stdout);
					zipper->i_map[row_in_real][col] = new_map;
					zipper->num_copy[row_in_real][col]++;

					if(j < zipper->m[i]-1)
					{
						row_diff = atoi(strtok(NULL, delims));
					}
				}
			}
			declared[4] = true;
		}
		else
		{
			printf("WARNING: cannot recognize option %s. Skipping this line\n", param);
		}
	}
	fclose(file);

	if(!(declared[0] && declared[1] && declared[2] && declared[3] && declared[4] && declared[5]))
	{
		printf("ERROR: at least one of period, m, n, k, interleaver_map, and buffer_len is missing in the configuration file\n");
		exit(1);
	}
}