/*
 * =====================================================================================
 *
 *       Filename:  common.h
 *
 *    Description:  function declarations for encoding and decoding of zipper codes
 *
 *        Version:  1.0
 *        Created:  2020-11-29 08:13:27 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Alvin Sukmadji 
 *   Organization:  University of Toronto
 *   
 *   Note: This license is a modified BSD License, which contains the terms of noncommercial
 *   use.
 *   
 *   Copyright <2020> <Alvin Yonathan Sukmadji>
 *   
 *   Redistribution and use in source and binary forms, with or without modification, are
 *   permitted provided that the following conditions are met:
 *   
 *   1. Redistributions of source code must retain the above copyright notice, this list of
 *   conditions and the following disclaimer.
 *   
 *   2. Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *   
 *   3. Neither the name of the copyright holder nor the names of its contributors may be used
 *   to endorse or promote products derived from this software without specific prior written
 *   permission.
 *   
 *   4. This software may only be used for noncommercial purposes, including teaching and
 *   academic research at universities, colleges and other educational institutions, and
 *   personal non-profit purposes. Neither this software, whether in Source or Object form,
 *   nor any derivative work that is based on (or derived from) this software may be used for
 *   commercial purposes, for example, licensing or selling the software, or using the
 *   software with a purpose to procure a commercial gain.
 *   
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * =====================================================================================
 */

#include "common.h"
#include "pcg_variants.h"

#ifndef ZIPPER_H
#define ZIPPER_H

// setting up parameters of zipper code by parsing the configuration file
// and allocating memory for the zipper struct
// inputs:
//     zipper    : pointer to the zipper code struct
//     filename  : filename (string) of the configuration file
//     chunk_size: the number of rows in a chunk, i.e., 'packet'
void zipper_setup(ZIPPER *zipper, char *filename, int chunk_size);

// encoding the latest chunk and transmitting it to the receiver
// assumes transmission over the binary symmetric channel with crossover probability error_prob
// inputs:
//     zipper          : pointer to the zipper code struct
//     newest_chunk_id : number of chunks transmitted so far
//     rng             : pointer to the ZMG (or PCG) random number generator
//     error_prob      : crossover probability over the BSC 
void zipper_encode(ZIPPER *zipper, long long int newest_chunk_id, pcg32f_random_t *rng, double error_prob);

// decoding codewords stored in the decoding window via iterative decoding
// returns the number of erroneous symbols in the oldest chunk of the window
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
//     num_info_symbols : pointer to the number of information symbols collected so far
unsigned int zipper_decode(ZIPPER *zipper, long long int newest_chunk_id, long long int *num_info_symbols);

// frees up memory allocated in zipper_setup
// inputs:
//     zipper : pointer to the zipper code struct
void zipper_cleanup(ZIPPER *zipper);

// prints the symbols in the transmitter-side buffer
// the number of bits printed out will be as many as the number of bits in the
// decoding window, with chunk no. newest_chunk_id being the newest (i.e., last) one
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
void zipper_print_buffer_sent(ZIPPER *zipper, long long int newest_chunk_id);

// prints the symbols in the receiver-side buffer
// the number of bits printed out will be as many as the number of bits in the
// decoding window, with chunk no. newest_chunk_id being the newest (i.e., last) one
// inputs:
//     zipper           : pointer to the zipper code struct
//     newest_chunk_id  : number of chunks transmitted so far
void zipper_print_buffer_recv(ZIPPER *zipper, long long int newest_chunk_id);

#endif //ZIPPER_H